/**
 * Copyright (c) 2009, 2012 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.activerecord;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.jdorm.activerecord.exceptions.ConnectionException;
import com.jdorm.datasource.DatabaseColumn;
import com.jdorm.datasource.DatabaseConnectionManager;
import com.jdorm.datasource.DatabaseDescriptor;
import com.jdorm.datasource.DatabaseRow;
import com.jdorm.datasource.GeneratedKey;
import com.jdorm.datasource.QueryAdministrator;
import com.jdorm.datasource.QueryResult;
import com.jdorm.inflection.Inflector;
import com.jdorm.logging.JDORMLogger;

/**
 * Base class that provides mapping of values in a database table
 * to Java objects. This class provides functionality to automatically
 * relate objects based on conventional foreign keys.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public abstract class AbstractActiveRecord implements Serializable {

	private static final long serialVersionUID = -362108770672310864L;
	
	public static final String CREATED_AT = "created_at"; //$NON-NLS-1$
	public static final String UPDATED_AT = "updated_at"; //$NON-NLS-1$
	
	private static final String ID = "id"; //$NON-NLS-1$
	private static final String EMPTY_STR = ""; //$NON-NLS-1$
	private static final String SELECT_ALL_FROM = "SELECT * FROM "; //$NON-NLS-1$
	private static final String AND = " AND "; //$NON-NLS-1$
	private static final char SPACE_CHAR = ' ';
	
	private static transient DatabaseDescriptor descriptor;
	private static transient QueryAdministrator administrator;
	
	private Map<String, Object> columnToDataMap;
	
	/**
	 * Initializes the ActiveRecord object.
	 */
	public AbstractActiveRecord() {
		columnToDataMap = new HashMap<String, Object>();
	}
	
	/**
	 * Establishes a connection to the database for all AbstractActiveRecord classes.
	 * 
	 * @param	descriptor	The descriptor to use to connect to the specified database.
	 * 
	 * @return	<code>true</code> if a connection was established,
	 * 			<code>false</code> otherwise.
	 */
	public static boolean establishConnection(DatabaseDescriptor descriptor) {
		AbstractActiveRecord.descriptor = descriptor;
		AbstractActiveRecord.administrator = null;
		Connection connection = null;
		try {
			connection = getConnection();
		} catch (ConnectionException e) {
			JDORMLogger.getLogger(AbstractActiveRecord.class).error(e);
		}
		return connection != null;
	}
	
	/**
	 * Gets a connection to the database for the current descriptor specified by
	 * the establishConnection method.
	 * 
	 * @return	A Connection to the SQL database.
	 * 
	 * @throws	ConnectionException	If a connection could not be established.
	 */
	protected static Connection getConnection() {
		if (AbstractActiveRecord.descriptor == null) {
			throw new ConnectionException(ActiveRecordMessages.AbstractActiveRecord_NoConnectionEstablished);
		}
		Connection connection = DatabaseConnectionManager.getInstance().getConnection(AbstractActiveRecord.descriptor);
		if (connection == null) {
			throw new ConnectionException(MessageFormat.format(ActiveRecordMessages.AbstractActiveRecord_CouldNotConnect, AbstractActiveRecord.descriptor.toString()));
		}
		return connection;
	}
	
	/**
	 * Gets an instance of the specified class. The specified class must
	 * provide a constructor with no parameters.
	 * 
	 * @param	clazz	The class to get an instance of.
	 * 
	 * @return	An instance of the type specified by the class.
	 */
	protected static <T> T getInstance(Class<T> clazz) {
		T t = null;
		try {
			t = clazz.newInstance();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		return t;
	}
	
	/**
	 * Gets the QueryAdministrator for the DatabaseDescriptor specified when calling
	 * establishConnection.
	 * 
	 * @return	The QueryAdministrator to use to execute SQL queries.
	 */
	protected static QueryAdministrator getQueryAdministrator() {
		if (AbstractActiveRecord.descriptor == null) {
			throw new ConnectionException(ActiveRecordMessages.AbstractActiveRecord_NoConnectionEstablished);
		}
		if (administrator == null) {
			administrator = new QueryAdministrator(descriptor);
		}
		return administrator;
	}
	
	/**
	 * Sets a value for the specified column.
	 * 
	 * @param	column	The column name for the object that is being specified.
	 * @param	value	The value to be stored at the specified column.
	 */
	public void set(String column, Object value) {
		columnToDataMap.put(column, value);
	}
	
	/**
	 * Gets the value for the specified column.
	 * 
	 * @param 	column	The column in the database.
	 * 
	 * @return	The value of the column.
	 */
	public Object get(String column) {
		return columnToDataMap.get(column);
	}
	
	/**
	 * Gets the string representation of the specified column.
	 * 
	 * @param	column	The column in the database.
	 * 
	 * @return	The string value of the column.
	 */
	public String getString(String column) {
		Object columnValue = get(column);
		return columnValue == null ? EMPTY_STR : columnValue.toString();
	}
	
	/**
	 * Gets the integer representation of the specified column.
	 * 
	 * @param	column	The column in the database.
	 * 
	 * @return	The integer value of the column.
	 */
	public Integer getInteger(String column) {
		String columnValue = getString(column);
		return columnValue.equals(EMPTY_STR) ? Integer.valueOf(0) : Integer.valueOf(columnValue);
	}
	
	/**
	 * Gets the boolean representation of the specified column.
	 * 
	 * @param	column	The column in the database.
	 * 
	 * @return	The boolean value of the column.
	 */
	public Boolean getBoolean(String column) {
		Boolean bool = null;
		if (get(column) instanceof Boolean) {
			bool = (Boolean) get(column);
		} else {
			String str = getString(column);
			try {
				int intVal = Integer.parseInt(str);
				bool = intVal == 1 ? Boolean.TRUE : Boolean.FALSE;
			} catch (NumberFormatException e) {
				// Could not be parsed, see if the string is "true"
				bool = Boolean.valueOf(str);
			}
		}
		return bool;
	}
	
	/**
	 * Gets the float representation of the specified column.
	 * 
	 * @param	column	The column in the database.
	 * 
	 * @return	The float value of the column.
	 */
	public Float getFloat(String column) {
		String columnValue = getString(column);
		return columnValue.equals(EMPTY_STR) ? Float.valueOf(0.0f) : Float.valueOf(columnValue);
	}
	
	/**
	 * Gets the double representation of the specified column.
	 * 
	 * @param	column	The column in the database.
	 * 
	 * @return	The double value of the column.
	 */
	public Double getDouble(String column) {
		String columnValue = getString(column);
		return columnValue.equals(EMPTY_STR) ? Double.valueOf(0.0d) : Double.valueOf(columnValue);
	}
	
	/**
	 * Gets the long representation of the specified column.
	 * 
	 * @param	column	The column in the database.
	 * 
	 * @return	The long value of the column.
	 */
	public Long getLong(String column) {
		String columnValue = getString(column);
		return columnValue.equals(EMPTY_STR) ? Long.valueOf(0L) : Long.valueOf(columnValue);
	}
	
	/**
	 * Gets the list of foreign relationships that have a foreign key pointing
	 * to the primary key of this object.
	 * 
	 * @param	clazz	The foreign object's class.
	 * 
	 * @return	A List of objects of the specified class that belong to this object. 
	 */
	public <T extends AbstractActiveRecord> List<T> getHasMany(Class<T> clazz) {
		return getHasMany(clazz, null);
	}
	
	/**
	 * Gets the list of foreign relationships that have a foreign key pointing
	 * to the primary key of this object.
	 * 
	 * @param	clazz	The foreign object's class.
	 * @param	conditions	The conditions to filter the results.
	 * 
	 * @return	A List of objects of the specified class that belong to this object. 
	 */
	public <T extends AbstractActiveRecord> List<T> getHasMany(Class<T> clazz, Conditions conditions) {
		return getHasMany(clazz, conditions, Inflector.getInstance().singularize(getDatabaseTableName(getClass())) + '_' + ID);
	}
	
	/**
	 * Gets the list of foreign relationships that have a foreign key pointing
	 * to the primary key of this object.
	 * 
	 * @param	clazz	The foreign object's class.
	 * @param	conditions	The conditions to filter the results.
	 * @param	foreignKey	The foreign key in the opposing table.
	 * 
	 * @return	A List of objects of the specified class that belong to this object.
	 */
	public <T extends AbstractActiveRecord> List<T> getHasMany(Class<T> clazz, Conditions conditions, String foreignKey) {
		String foreignKeyWhere = foreignKey + " = ?"; //$NON-NLS-1$
		if (conditions == null) {
			conditions = Conditions.where(foreignKeyWhere, get(ID));
		} else {
			if (conditions.getWhere() != null) {
				Object[] origParameters = conditions.getParameters();
				Object[] params = Arrays.copyOf(origParameters, origParameters.length + 1);
				params[origParameters.length] = get(ID);
				conditions.setWhere(conditions.getWhere() + AND + foreignKeyWhere, params);
			} else {
				conditions.setWhere(foreignKeyWhere, get(ID));
			}
		}
		QueryResult results = getQueryAdministrator().executeAndGetResult(SELECT_ALL_FROM + getDatabaseTableName(clazz) + SPACE_CHAR + conditions.toString(), conditions.getParameters());
		return getRecordsForResult(results, clazz);
	}
	
	/**
	 * Gets the list of foreign relationships that have a foreign key pointing
	 * to the primary key of this object through an intermediate table.
	 * 
	 * @param	clazz	The foreign object's class.
	 * @param	through	The through object's class.
	 * 
	 * @return	A List of objects of the specified class that belong to this object.
	 */
	public <T extends AbstractActiveRecord> List<T> getHasManyThrough(Class<T> clazz, Class<? extends AbstractActiveRecord> through) {
		return getHasManyThrough(clazz, through, Inflector.getInstance().singularize(getDatabaseTableName(clazz)) + '_' + ID, Inflector.getInstance().singularize(getDatabaseTableName(getClass())) + '_' + ID);
	}
	
	/**
	 * Gets the list of foreign relationships that have a foreign key pointing
	 * to the primary key of this object through an intermediate table.
	 * 
	 * @param	clazz	The foreign object's class.
	 * @param	through	The through object's class.
	 * @param	foreignKey	The foreign key in the intermediate table for the foreign object.
	 * @param	throughKey 	The foreign key in the intermediate table for the local object.
	 * 
	 * @return	A List of objects of the specified class that belong to this object.
	 */
	public <T extends AbstractActiveRecord> List<T> getHasManyThrough(Class<T> clazz, Class<? extends AbstractActiveRecord> through, String foreignKey, String throughKey) {
		QueryResult results = getQueryAdministrator().executeAndGetResult("SELECT a.* FROM " + getDatabaseTableName(clazz) + " a, " + getDatabaseTableName(through) + " b WHERE b." + throughKey + " = ? AND a.id = b." + foreignKey, get(ID)); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		return getRecordsForResult(results, clazz);
	}
	
	/**
	 * Gets the foreign object that this object relates to based on the
	 * specified class.
	 * 
	 * @param	clazz	The class that this object points to.
	 * 
	 * @return	The object that this class points to.
	 */
	public <T extends AbstractActiveRecord> T getBelongsTo(Class<T> clazz) {
		return getBelongsTo(clazz, null);
	}
	
	/**
	 * Gets the foreign object that this object relates to based on the
	 * specified class.
	 * 
	 * @param	clazz	The class that this object points to.
	 * @param	conditions	The conditions to filter the results.
	 * 
	 * @return	The object that this class points to.
	 */
	public <T extends AbstractActiveRecord> T getBelongsTo(Class<T> clazz, Conditions conditions) {
		return getBelongsTo(clazz, conditions, Inflector.getInstance().singularize(getDatabaseTableName(clazz)) + '_' + ID);
	}
	
	/**
	 * Gets the foreign object that this object relates to based on the
	 * specified class and foreign key.
	 * 
	 * @param	clazz	The foreign object's class.
	 * @param	conditions	The conditions to filter the results.
	 * @param	foreignKey	The foreign key in this table pointing to the foreign object.
	 * 
	 * @return	The object that this class points to.
	 */
	public <T extends AbstractActiveRecord> T getBelongsTo(Class<T> clazz, Conditions conditions, String foreignKey) {
		Object foreignKeyId = get(foreignKey);
		String foreignKeyWhere = ID + " = ?"; //$NON-NLS-1$
		if (conditions == null) {
			conditions = Conditions.where(foreignKeyWhere, foreignKeyId);
		} else {
			if (conditions.getWhere() != null) {
				Object[] origParameters = conditions.getParameters();
				Object[] params = Arrays.copyOf(origParameters, origParameters.length + 1);
				params[origParameters.length] = foreignKeyId;
				conditions.setWhere(conditions.getWhere() + AND + foreignKeyWhere, params);
			} else {
				conditions.setWhere(foreignKeyWhere, foreignKeyId);
			}
		}
		QueryResult results = getQueryAdministrator().executeAndGetResult(SELECT_ALL_FROM + getDatabaseTableName(clazz) + SPACE_CHAR + conditions.toString(), 1, conditions.getParameters());
		List<T> records = getRecordsForResult(results, clazz);
		T record = null;
		if (records.size() > 0) {
			record = records.get(0);
		}
		return record;
	}
	
	/**
	 * Finds all of the objects of the specified type.
	 * 
	 * @param	clazz	The type to find all of the objects for.
	 * 
	 * @return	A List of all the objects of the specified class.
	 */
	public static <T extends AbstractActiveRecord> List<T> findAll(Class<T> clazz) {
		return find(clazz, QueryAdministrator.UNLIMITED_FETCH);
	}
	
	/**
	 * Finds all of the objects of the specified type.
	 * 
	 * @param	clazz	The type to find all of the objects for.
	 * @param	conditions	The conditions for the find.
	 * 
	 * @return	A List of all the objects of the specified class.
	 */
	public static <T extends AbstractActiveRecord> List<T> findAll(Class<T> clazz, Conditions conditions) {
		return find(clazz, conditions, QueryAdministrator.UNLIMITED_FETCH);
	}
	
	/**
	 * Finds the first object of the specified type.
	 * 
	 * @param	clazz	The type to find the object for.
	 * @param	conditions	The conditions for the find.
	 * 
	 * @return	The first object of the specified class.
	 */
	public static <T extends AbstractActiveRecord> T findFirst(Class<T> clazz, Conditions conditions) {
		List<T> records = find(clazz, conditions, 1);
		T record = null;
		if (records.size() > 0) {
			 record = records.get(0);
		}
		return record;
	}
	
	/**
	 * Finds the first object of the specified type.
	 * 
	 * @param	clazz	The type to find the object for.
	 * 
	 * @return	The first object of the specified class.
	 */
	public static <T extends AbstractActiveRecord> T findFirst(Class<T> clazz) {
		return findFirst(clazz, null);
	}
	
	/**
	 * Finds a list of objects of the specified type.
	 * 
	 * @param	clazz	The type to find the objects for.
	 * @param	fetchSize	The number of objects to return.
	 * 
	 * @return	A List of all the objects of the specified class, limited to the specified fetch size.
	 */
	public static <T extends AbstractActiveRecord> List<T> find(Class<T> clazz, int fetchSize) {
		return find(clazz, null, fetchSize);
	}
	
	/**
	 * Finds a list of objects of the specified type with the specified where clause.
	 * 
	 * @param	clazz	The type to find the objects for.
	 * @param	conditions	The conditions for the find.
	 * @param	fetchSize	The number of objects to return.
	 * 
	 * @return	A List of all the objects of the specified class, limited to the specified fetch size and where condition.
	 */
	public static <T extends AbstractActiveRecord> List<T> find(Class<T> clazz, Conditions conditions, int fetchSize) {
		String conditionsStr = EMPTY_STR;
		Object[] parameters = null;
		if (conditions != null) {
			conditionsStr = SPACE_CHAR + conditions.toString();
			parameters = conditions.getParameters();
		}
		QueryResult results = getQueryAdministrator().executeAndGetResult(SELECT_ALL_FROM + getDatabaseTableName(clazz) + conditionsStr, fetchSize, parameters);
		return getRecordsForResult(results, clazz);
	}
	
	/**
	 * Gets a list of objects of the specified class from the QueryResult.
	 * 
	 * @param	results	The QueryResult to get the objects from.
	 * @param	clazz	The type of objects to return.
	 * 
	 * @return	A List of objects of the specified type.
	 */
	private static <T extends AbstractActiveRecord> List<T> getRecordsForResult(QueryResult results, Class<T> clazz) {
		List<T> records = new ArrayList<T>();
		for (DatabaseRow row : results.getRows()) {
			T instance = getInstance(clazz);
			for (int i = 0; i < results.getColumnCount(); i++) {
				instance.set(results.getColumns().get(i).getName(), row.getData(i));
			}
			records.add(instance);
		}
		return records;
	}
	
	/**
	 * Gets the name of the database table.
	 * 
	 * @param	clazz	The class to get the table name for.
	 * 
	 * @return	The name of the database table.
	 */
	public static <T extends AbstractActiveRecord> String getDatabaseTableName(Class<T> clazz) {
		StringBuilder builder = new StringBuilder();
		String className = clazz.getSimpleName();
		for (int i = 0; i < className.length(); i++) {
			String originalCharacter = String.valueOf(className.charAt(i));
			String lowercase = originalCharacter.toLowerCase();
			if (!originalCharacter.equals(lowercase) && i != 0) {
				builder.append('_');
			}
			builder.append(lowercase);
		}
		return Inflector.getInstance().pluralize(builder.toString());
	}
	
	/**
	 * Gets a list of the fields associated with this ActiveRecord.
	 * 
	 * @param	clazz	The class to get the fields for.
	 * 
	 * @return	A list of the fields/columns for this ActiveRecord instance.
	 */
	public static <T extends AbstractActiveRecord> List<String> getFields(Class<T> clazz) {
		List<String> fields = new LinkedList<String>();
		for (DatabaseColumn column : getQueryAdministrator().getColumns(getDatabaseTableName(clazz))) {
			fields.add(column.getName());
		}
		return fields;
	}
	
	/**
	 * Determines whether this record is a new, unsaved record.
	 * 
	 * @return	<code>true</code> if this record is new and isn't saved in the database,
	 * 			<code>false</code> otherwise.
	 */
	public boolean isNewRecord() {
		return getInteger(ID) == Integer.valueOf(0);
	}
	
	/**
	 * Saves this record to the database. If this is a new record, the record will be
	 * inserted into the database, otherwise it will be updated in the database.
	 * 
	 * @return	<code>true</code> if the record was saved successfully,
	 * 			<code>false</code> otherwise.
	 */
	public boolean save() {
		boolean saved = false;
		// Attempt to find if we have created_at and updated_at
		if (isNewRecord()) {
			updateTimestamps(CREATED_AT, UPDATED_AT);
			saved = insertRecord();
		} else {
			updateTimestamps(UPDATED_AT);
			saved = updateRecord();
		}
		return saved;
	}
	
	/**
	 * Updates the specified timestamp fields to the current time.
	 * 
	 * @param	fields	The list of fields to update.
	 */
	private void updateTimestamps(String... fields) {
		long currentTime = Calendar.getInstance().getTimeInMillis();
		for (DatabaseColumn column : getQueryAdministrator().getColumns(getDatabaseTableName(getClass()))) {
			if (column.getType() == Types.TIMESTAMP) {
				for (int i = 0; i < fields.length; i++) {
					if (fields[i].equals(column.getName())) {
						columnToDataMap.put(column.getName(), new Timestamp(currentTime));
					}
				}
			}
		}
	}
	
	/**
	 * Inserts this record to the database.
	 * 
	 * @return	<code>true</code> if the record was inserted successfully,
	 * 			<code>false</code> otherwise.
	 */
	private boolean insertRecord() {
		// The start of the SQL query
        StringBuilder sql = new StringBuilder("INSERT INTO ").append(getDatabaseTableName(getClass())) //$NON-NLS-1$
        		.append(" ("); //$NON-NLS-1$
        
        Object[] parameters = new Object[columnToDataMap.size()];
        StringBuilder values = new StringBuilder();
        int count = 0;
        for (String column : columnToDataMap.keySet()) {
        	// If this isn't the first column, separate with a comma
        	if (count > 0) {
                sql.append(", "); //$NON-NLS-1$
                values.append(", "); //$NON-NLS-1$
            }
            sql.append(column);
            values.append("?"); //$NON-NLS-1$
            parameters[count++] = columnToDataMap.get(column);
        }
        
        sql.append(") VALUES (").append(values).append(")"); //$NON-NLS-1$ //$NON-NLS-2$
        
		// Execute the SQL
        boolean success = getQueryAdministrator().execute(sql.toString(), parameters);
		if (success) {
			// Attempt to get the ID
			List<GeneratedKey> keys = getQueryAdministrator().getGeneratedKeys();
			if (!keys.isEmpty()) {
				Object id = keys.get(0).getData();
				if (id != null) {
					set(ID, id);
				}
			}
		}
		return success;
	}
	
	/**
	 * Updates this record in the database.
	 * 
	 * @return	<code>true</code> if the record was updated successfully,
	 * 			<code>false</code> otherwise.
	 */
	private boolean updateRecord() {
		StringBuilder sql = new StringBuilder("UPDATE ").append(getDatabaseTableName(getClass())) //$NON-NLS-1$
				.append(" SET "); //$NON-NLS-1$

		// The additional 1 is for the where clause
		Object[] parameters = new Object[columnToDataMap.size() + 1];
		int count = 0;
		long currentTime = Calendar.getInstance().getTimeInMillis();
        for (String column : columnToDataMap.keySet()) {
        	// If this isn't the first column, separate with a comma
        	if (count > 0) {
                sql.append(", "); //$NON-NLS-1$
            }
            Object value = null;
            if (column.equals(UPDATED_AT)) {
        		value = new Timestamp(currentTime);
            } else {
            	value = get(column);
            }
            sql.append(column).append(" = ?"); //$NON-NLS-1$
            parameters[count++] = value;
        }
        sql.append(" WHERE id = ?"); //$NON-NLS-1$
        parameters[count] = get(ID);
        return getQueryAdministrator().execute(sql.toString(), parameters);
	}
	
	/**
     * Removes the object from the database.
     *
     * @return  <code>true</code> if the object was successfully destroyed,
     *          <code>false</code> otherwise.
     */
	public boolean destroy() {
		boolean success = false;
		if (!isNewRecord()) {
			StringBuilder sql = new StringBuilder("DELETE FROM ") //$NON-NLS-1$
					.append(getDatabaseTableName(getClass())).append(" WHERE id = ?"); //$NON-NLS-1$
	        success = getQueryAdministrator().execute(sql.toString(), get(ID));
	        if (success) {
	        	columnToDataMap.remove(ID);
	        }
		}
        return success;
	}
	
	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		boolean retVal = false;
		Class<? extends AbstractActiveRecord> clazz = getClass();
		if (clazz.isInstance(obj)) {
			AbstractActiveRecord record = clazz.cast(obj);
			retVal = getInteger(ID).equals(record.getInteger(ID));
		}
		return retVal;
	}
	
} // AbstractActiveRecord