/**
 * Copyright (c) 2009, 2011 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.activerecord;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 * Represents conditions for an active record find. This class
 * is designed to allow chaining the methods, for example:
 * 
 * Conditions.where("username = ?", "mbaumbach").setOrder("last_name DESC");
 * 
 * Question marks are markers for variable binding. These variables are
 * automatically escaped so as to reduce the possibility of introducing
 * queries that could contain SQL injection security holes.
 * 
 * Any part of a query that is dynamically added via a variable that is
 * accessible to modification by an end user should use variable binding.
 * 
 * In addition to security, this will automatically escape characters that
 * could cause an SQL statement to fail.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class Conditions {
	
	private String where;
	private Object[] whereParameters;
	private String order;
	private Object[] orderParameters;
	private String group;
	private Object[] groupParameters;
	private String having;
	private Object[] havingParameters;
	
	/**
	 * Private constructor, use the static methods to create conditions.
	 */
	private Conditions() {}
	
	/**
	 * Sets the where condition of this conditions object.
	 * 
	 * @param	where	The WHERE condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public static Conditions where(String where, Object... parameters) {
		return new Conditions().setWhere(where, parameters);
	}
	
	/**
	 * Sets the order condition of this conditions object.
	 * 
	 * @param	order	The ORDER BY condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public static Conditions order(String order, Object... parameters) {
		return new Conditions().setOrder(order, parameters);
	}
	
	/**
	 * Sets the group condition of this conditions object.
	 * 
	 * @param	group	The GROUP BY condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public static Conditions group(String group, Object... parameters) {
		return new Conditions().setGroup(group, parameters);
	}
	
	/**
	 * Sets the having condition of this conditions object.
	 * 
	 * @param	having	The HAVING condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public static Conditions having(String having, Object... parameters) {
		return new Conditions().setHaving(having, parameters);
	}
	
	/**
	 * Sets the where condition of this conditions object.
	 * 
	 * @param	where	The WHERE condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public Conditions setWhere(String where, Object... parameters) {
		this.where = where;
		whereParameters = Arrays.copyOf(parameters, parameters.length);
		return this;
	}
	
	/**
	 * Sets the order condition of this conditions object.
	 * 
	 * @param	order	The ORDER BY condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public Conditions setOrder(String order, Object... parameters) {
		this.order = order;
		orderParameters = Arrays.copyOf(parameters, parameters.length);
		return this;
	}
	
	/**
	 * Sets the group condition of this conditions object.
	 * 
	 * @param	group	The GROUP BY condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public Conditions setGroup(String group, Object... parameters) {
		this.group = group;
		groupParameters = Arrays.copyOf(parameters, parameters.length);
		return this;
	}
	
	/**
	 * Sets the having condition of this conditions object.
	 * 
	 * @param	having	The HAVING condition of an SQL statement.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	The conditions object this was called from, allowing
	 * 			this method to be chained.
	 */
	public Conditions setHaving(String having, Object... parameters) {
		this.having = having;
		havingParameters = Arrays.copyOf(parameters, parameters.length);
		return this;
	}
	
	/**
	 * Gets the where condition.
	 * 
	 * @return	The where condition.
	 */
	public String getWhere() {
		return where;
	}

	/**
	 * Gets the order condition.
	 * 
	 * @return	The order condition.
	 */
	public String getOrder() {
		return order;
	}

	/**
	 * Gets the group condition.
	 * 
	 * @return	The group condition.
	 */
	public String getGroup() {
		return group;
	}

	/**
	 * Gets the having condition.
	 * 
	 * @return	The having condition.
	 */
	public String getHaving() {
		return having;
	}

	/**
	 * Get all of the parameters for this conditions that will be bound to question marks
	 * in the SQL query.
	 * 
	 * @return	An array of object parameters that will be bound to the query.
	 */
	public Object[] getParameters() {
		List<Object> params = new LinkedList<Object>();
		if (getWhere() != null && !getWhere().trim().isEmpty()) {
			addParameters(params, whereParameters);
		}
		if (getOrder() != null && !getOrder().trim().isEmpty()) {
			addParameters(params, orderParameters);
		}
		if (getGroup() != null && !getGroup().trim().isEmpty()) {
			addParameters(params, groupParameters);
			if (getHaving() != null && !getHaving().trim().isEmpty()) {
				addParameters(params, havingParameters);	
			}
		}
		return params.toArray();
	}
	
	/**
	 * Helper method for adding parameters to a list if paramsToAdd is not null.
	 * 
	 * @param	params	The list to hold all of the parameters.
	 * @param 	paramsToAdd	The parameters to add. If null, nothing in params will be changed.
	 */
	private void addParameters(List<Object> params, Object[] paramsToAdd) {
		if (paramsToAdd != null) {
			for (Object o : paramsToAdd) {
				params.add(o);
			}
		}
	}
	
	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuilder conditions = new StringBuilder();
		if (getWhere() != null && !getWhere().trim().isEmpty()) {
			conditions.append("WHERE ").append(getWhere().trim()); //$NON-NLS-1$
		}
		if (getOrder() != null && !getOrder().trim().isEmpty()) {
			if (conditions.length() > 0) {
				conditions.append(' ');
			}
			conditions.append("ORDER BY ").append(getOrder().trim()); //$NON-NLS-1$
		}
		if (getGroup() != null && !getGroup().trim().isEmpty()) {
			if (conditions.length() > 0) {
				conditions.append(' ');
			}
			conditions.append("GROUP BY ").append(getGroup().trim()); //$NON-NLS-1$
			if (getHaving() != null && !getHaving().trim().isEmpty()) {
				conditions.append(' ').append("HAVING ").append(getHaving().trim()); //$NON-NLS-1$
			}
		}
		return conditions.toString();
	}
	
	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object o) {
		boolean retVal = false;
		if (o instanceof Conditions) {
			retVal = Conditions.buildObjectString(this).equals(Conditions.buildObjectString((Conditions) o));
		}
		return retVal;
	}
	
	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return Conditions.buildObjectString(this).hashCode();
	}
	
	/**
	 * Helper method for creating a string that represents both the SQL query
	 * and the parameters that are found to the query. Regardless of what order
	 * the conditions are built in, this string will always be the same as long
	 * as the individual calls are equivalent.
	 * 
	 * @param	conditions	The conditions object to build the string for.
	 * 
	 * @return	A string that can be used for comparisons of conditions clauses.
	 */
	private static String buildObjectString(Conditions conditions) {
		StringBuilder builder = new StringBuilder(conditions.toString());
		builder.append(" (Parameters: "); //$NON-NLS-1$
		Object[] params = conditions.getParameters();
		for (int i = 0; i < params.length; i++) {
			if (i > 0) {
				builder.append(", "); //$NON-NLS-1$
			}
			builder.append(params[i].toString());
		}
		builder.append(")"); //$NON-NLS-1$
		return builder.toString();
	}
	
} // Conditions