/**
 * Copyright (c) 2009, 2010 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.datasource;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.sql.DataSource;

import com.jdorm.datasource.connectors.AbstractDataSourceConnector;
import com.jdorm.logging.JDORMLogger;

/**
 * Manages all connections to databases. This manager will attempt to reuse
 * connections to the database when possible and will automatically create
 * a connection to the database should the connection ever close.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class DatabaseConnectionManager {
	
	private static JDORMLogger logger = JDORMLogger.getLogger(DatabaseConnectionManager.class);
	private static DatabaseConnectionManager instance;
	
	private Map<DatabaseDescriptor, Connection> descriptorToConnection;
	
	/**
	 * Private constructor since this class is a singleton. Use the getInstance()
	 * method to get the only instance of this class.
	 */
	private DatabaseConnectionManager() {
		descriptorToConnection = new HashMap<DatabaseDescriptor, Connection>();
	}
	
	/**
	 * Gets the only instance of this class.
	 * 
	 * @return	The instance of DatabaseConnectionManager.
	 */
	public static DatabaseConnectionManager getInstance() {
		if (instance == null) {
			instance = new DatabaseConnectionManager();
		}
		return instance;
	}
	
	/**
	 * Gets a connection for the specified database descriptor.
	 * 
	 * @param	descriptor	The descriptor defining the connection information.
	 * 
	 * @return	A Connection to the database, or <code>null</code> if one could not be created.
	 */
	@SuppressWarnings("resource")
	public Connection getConnection(DatabaseDescriptor descriptor) {
		Connection connection = descriptorToConnection.get(descriptor);
		try {
			if (connection == null || (connection != null && !connection.isValid(0))) {
				if (descriptor.getType() == DataSourceType.JDBCDataSource) {
					Context env = new InitialContext();
		            DataSource pool = (DataSource) env.lookup("java:comp/env/jdbc/" + descriptor.getDatabase()); //$NON-NLS-1$
		            if (pool != null) {
						if (descriptor.hasUsername()) {
							connection = pool.getConnection(descriptor.getUsername(), descriptor.getPassword());
						} else {
							connection = pool.getConnection();
						}
		            }
				} else {
					// There is no valid connection, attempt to connect now
					AbstractDataSourceConnector connector = descriptor.getType().getConnector();
					Class.forName(connector.getDriver());
					String host = descriptor.hasHost() ? descriptor.getHost() : AbstractDataSourceConnector.DEFAULT_HOST;
					String port = String.valueOf(descriptor.hasPort() ? descriptor.getPort() : connector.getDefaultPort());
					String url = connector.getUrl(host, port, descriptor.getDatabase());
					if (descriptor.hasUsername()) {
						connection = DriverManager.getConnection(url, descriptor.getUsername(), descriptor.getPassword());
					} else {
						connection = DriverManager.getConnection(url);
					}
				}
				descriptorToConnection.put(descriptor, connection);
			}
		} catch (Exception e) {
			logger.error(DataSourceMessages.DatabaseConnectionManager_CannotConnect, e);
		}
		return connection;
	}
	
	/**
	 * Forces the connection associated with the database descriptor to be closed.
	 * 
	 * @param	descriptor	The descriptor to attempt to close.
	 * 
	 * @return	<code>true</code> if the connection did not exist, was already closed,
	 * 			or was closed successfully. <code>false</code> otherwise.
	 */
	public boolean closeConnection(DatabaseDescriptor descriptor) {
		boolean retVal = false;
		Connection connection = descriptorToConnection.get(descriptor);
		try {
			if (connection != null && !connection.isClosed()) {
				connection.close();
			}
			retVal = true;
		} catch (SQLException e) {
			logger.error(DataSourceMessages.DatabaseConnectionManager_CannotCloseConnection, e);
		}
		return retVal;
	}

} // DatabaseConnectionManager