/**
 * Copyright (c) 2009, 2010 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.datasource;

import com.jdorm.datasource.connectors.AbstractDataSourceConnector;

/**
 * Describes the connection information needed to connect
 * to a specific database.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class DatabaseDescriptor {

	private DataSourceType type;
	private String database;
	private String username;
	private String password;
	private String host;
	private int port;
	
	/**
	 * Creates a database descriptor.
	 * 
	 * @param	type	The DataSourceType specifying the type of data source.
	 * @param	database	The database to connect to.
	 * @param	username	The username to connect with.
	 * @param	password	The password to connect with.
	 * @param	host	The host to connect to.
	 * @param	port	The port to connect through.
	 */
	public DatabaseDescriptor(DataSourceType type, String database, String username, String password, String host, int port) {
		this.type = type;
		this.database = database;
		this.username = username;
		this.password = password;
		this.host = host;
		this.port = port;
	}
	
	/**
	 * Creates a database descriptor.
	 * 
	 * @param	type	The DataSourceType specifying the type of data source.
	 * @param	database	The database to connect to.
	 * @param	username	The username to connect with.
	 * @param	password	The password to connect with.
	 */
	public DatabaseDescriptor(DataSourceType type, String database, String username, String password) {
		this(type, database, username, password, null, -1);
	}
	
	/**
	 * Creates a database descriptor.
	 * 
	 * @param	type	The DataSourceType specifying the type of data source.
	 * @param	database	The database to connect to.
	 */
	public DatabaseDescriptor(DataSourceType type, String database) {
		this(type, database, null, null);
	}

	/**
	 * Gets the type.
	 *
	 * @return	The type.
	 */
	public DataSourceType getType() {
		return type;
	}

	/**
	 * Sets the type to type.
	 *
	 * @param	type	The type to set.
	 */
	public void setType(DataSourceType type) {
		this.type = type;
	}

	/**
	 * Gets the database.
	 *
	 * @return	The database.
	 */
	public String getDatabase() {
		return database;
	}

	/**
	 * Sets the database to database.
	 *
	 * @param	database	The database to set.
	 */
	public void setDatabase(String database) {
		this.database = database;
	}

	/**
	 * Gets the username.
	 *
	 * @return	The username.
	 */
	public String getUsername() {
		return username;
	}

	/**
	 * Sets the username to username.
	 *
	 * @param	username	The username to set.
	 */
	public void setUsername(String username) {
		this.username = username;
	}

	/**
	 * Gets the password.
	 *
	 * @return	The password.
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Sets the password to password.
	 *
	 * @param	password	The password to set.
	 */
	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * Gets the host.
	 *
	 * @return	The host.
	 */
	public String getHost() {
		return host;
	}

	/**
	 * Sets the host to host.
	 *
	 * @param	host	The host to set.
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * Gets the port.
	 *
	 * @return	The port.
	 */
	public int getPort() {
		return port;
	}

	/**
	 * Sets the port to port.
	 *
	 * @param	port	The port to set.
	 */
	public void setPort(int port) {
		this.port = port;
	}
	
	/**
	 * Determines whether this descriptor has a port set on it.
	 * 
	 * @return	<code>true</code> if there is a port specified,
	 *			<code>false</code> otherwise. 				
	 */
	public boolean hasPort() {
		return port != -1;
	}
	
	/**
	 * Determines whether this descriptor has a host set on it.
	 * 
	 * @return	<code>true</code> if there is a host specified,
	 *			<code>false</code> otherwise. 				
	 */
	public boolean hasHost() {
		return host != null;
	}
	
	/**
	 * Determines whether this descriptor has a username set on it.
	 * 
	 * @return	<code>true</code> if there is a username specified,
	 *			<code>false</code> otherwise. 				
	 */
	public boolean hasUsername() {
		return username != null;
	}
	
	/**
	 * Determines whether this descriptor has a password set on it.
	 * 
	 * @return	<code>true</code> if there is a password specified,
	 *			<code>false</code> otherwise. 				
	 */
	public boolean hasPassword() {
		return password != null;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		boolean retVal = false;
		if (obj instanceof DatabaseDescriptor) {
			DatabaseDescriptor o = (DatabaseDescriptor) obj;
			retVal = getType().equals(o.getType()) 
				&& ((getHost() == null && o.getHost() == null) || getHost().equals(o.getHost()))
				&& getPort() == o.getPort()
				&& ((getPassword() == null && o.getPassword() == null) || getPassword().equals(o.getPassword()));
		}
		return retVal;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		StringBuilder builder = new StringBuilder(getType().toString());
		builder.append(":").append(getHost()) //$NON-NLS-1$
			.append(":").append(String.valueOf(getPort())) //$NON-NLS-1$
			.append(":").append(getUsername()) //$NON-NLS-1$
			.append(":").append(getPassword()); //$NON-NLS-1$
		return builder.toString().hashCode();
	}
	
	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuffer buffer = new StringBuffer(hasHost() ? getHost() : AbstractDataSourceConnector.DEFAULT_HOST);
		buffer.append('\n').append(getPort())
			.append('\n').append(getType().toString())
			.append('\n').append(getDatabase())
			.append('\n').append(getUsername())
			.append('\n').append(getPassword());
		return buffer.toString();
	}
	
} // DatabaseDescriptor