/**
 * Copyright (c) 2009, 2011 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.datasource;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import com.jdorm.logging.JDORMLogger;

/**
 * Executes SQL statements on a JDBC Connection and returns
 * result sets in the form of vanilla ResultSet objects or
 * wrapped QueryResult objects.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class QueryAdministrator {
	
	/**
	 * Specifies fetching an unlimited number of rows.
	 */
	public static final int UNLIMITED_FETCH = 0;
	
	private static JDORMLogger logger = JDORMLogger.getLogger(QueryAdministrator.class);
	
	private DatabaseDescriptor descriptor;
	private QueryResult lastResult;
	private List<GeneratedKey> generatedKeys;
	
	/**
	 * Creates a new QueryAdministrator based on the DatabaseDescriptor.
	 * 
	 * @param	descriptor	The database descriptor used to connect to the database.
	 */
	public QueryAdministrator(DatabaseDescriptor descriptor) {
		this.descriptor = descriptor;
		generatedKeys = new LinkedList<GeneratedKey>();
	}
	
	/**
	 * Gets a reference to the DatabaseDescriptor used to create this object.
	 * 
	 * @return	The DatabaseDescriptor used to by this QueryAdministrator.
	 */
	public DatabaseDescriptor getDatabaseDescriptor() {
		return descriptor;
	}
	
	/**
	 * Gets a connection to the database.
	 * 
	 * @return	A Connection to the database.
	 */
	protected Connection getConnection() {
		return DatabaseConnectionManager.getInstance().getConnection(descriptor);
	}
	
	/**
	 * Executes a SQL statement on the connection.
	 * 
	 * @param	sql	The SQL statement to execute.
	 * @param	fetchSize	The number of rows to return.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	A ResultSet of the resulting data based on the execution of the statement.
	 * 
	 * @throws	SQLException	If an error occurred parsing or running the SQL statement.
	 */
	public ResultSet executeQuery(String sql, int fetchSize, Object... parameters) throws SQLException {
		ResultSet retVal = null;
		Connection conn = getConnection();
		if (conn != null) {
			PreparedStatement statement = null;
			statement = conn.prepareStatement(sql, PreparedStatement.RETURN_GENERATED_KEYS);
			if (parameters != null) {
				for (int i = 0; i < parameters.length; i++) {
					statement.setObject(i + 1, parameters[i]);
				}
			}
			statement.setFetchSize(fetchSize);
			statement.execute();
			setGeneratedKeys(statement);
			retVal = statement.getResultSet();
		}
		return retVal;
	}
	
	/**
	 * Executes a SQL statement on the connection.
	 * 
	 * @param	sql	The SQL statement to execute.
	 * @param	parameters	The parameters to fill into the SQL statement.
	 * 
	 * @return	A ResultSet of the resulting data based on the execution of the statement.
	 * 
	 * @throws	SQLException	If an error occurred parsing the SQL statement.
	 */
	public ResultSet executeQuery(String sql, Object... parameters) throws SQLException {
		return executeQuery(sql, QueryAdministrator.UNLIMITED_FETCH, parameters);
	}
	
	/**
     * Executes the specified query.
     *
     * @param   sql The SQL query to execute.
     * @param	fetchSize	The number of rows to return.
     * @param	parameters	The parameters to fill into the SQL statement.
     *
     * @return  <code>true</code> if the query was successful,
     *          <code>false</code> otherwise.
     */
    public boolean execute(String sql, int fetchSize, Object... parameters) {
    	boolean successful = false;
    	lastResult = null;
    	ResultSet rs = null;
    	try {
			rs = executeQuery(sql, fetchSize, parameters);
			if (rs != null) {
				lastResult = new QueryResult(rs);
			}
			successful = true;
		} catch (SQLException e) {
			logger.error(e);
		} finally {
			closeResultSet(rs);
		}
		return successful;
    }
	
    /**
     * Executes the specified query.
     *
     * @param   sql The SQL query to execute.
     * @param	parameters	The parameters to fill into the SQL statement.
     *
     * @return  <code>true</code> if the query was successful,
     *          <code>false</code> otherwise.
     */
    public boolean execute(String sql, Object... parameters) {
    	return execute(sql, QueryAdministrator.UNLIMITED_FETCH, parameters);
    }
    
    /**
     * Executes the specified query and returns the QueryResult.
     *
     * @param   sql The SQL query to execute.
     * @param	fetchSize	The number of rows to return.
     * @param	parameters	The parameters to fill into the SQL statement.
     * 
     * @return	The QueryResult created by executing the query,
     * 			or <code>null</code> if the query did not execute successfully.
     */
    public QueryResult executeAndGetResult(String sql, int fetchSize, Object... parameters) {
    	return execute(sql, fetchSize, parameters) ? getLastResult() : null;
    }
    
    /**
     * Executes the specified query and returns the QueryResult.
     *
     * @param   sql The SQL query to execute.
     * @param	parameters	The parameters to fill into the SQL statement.
     * 
     * @return	The QueryResult created by executing the query,
     * 			or <code>null</code> if the query did not execute successfully.
     */
    public QueryResult executeAndGetResult(String sql, Object... parameters) {
    	return executeAndGetResult(sql, QueryAdministrator.UNLIMITED_FETCH, parameters);
    }
    
	/**
	 * Gets the last result of a call to the execute method.
	 * 
	 * @return	A QueryResult object, or <code>null</code> if no query has
     *          been executed.
	 */
	public QueryResult getLastResult() {
		return lastResult;
	}
	
	/**
	 * This method checks if we're using Oracle and gets the correct schema
	 * pattern for JDBC.
	 * 
	 * @return	The correct schema pattern for Oracle (which is the Username),
	 * 			or <code>null</code> for most other connections
	 */
	private String getSchemaPattern() {
		Connection conn = getConnection();
		String retVal = null;
		if (conn != null && isOracle()) {
			try {
				retVal = conn.getMetaData().getUserName();
			} catch (SQLException e) {
				logger.error(e);
			}
		}
		return retVal;
	}
	
	/**
	 * Determines whether we are connected to an Oracle database.
	 * 
	 * @return	<code>true</code> if the database is Oracle,
	 * 			<code>false</code> otherwise.
	 */
	private boolean isOracle() {
		return descriptor.getType().equals(DataSourceType.Oracle);
	}
	
	/**
	 * Gets the list of tables for this database.
	 * 
	 * @return	A collection, represented in Strings, of the name of tables
	 * 			in the database.
	 */
	public Collection<String> getTables() {
		Collection<String> tables = new LinkedList<String>();
		Connection conn = getConnection();
		if (conn != null) {
			ResultSet rs = null;
			try {
				rs = conn.getMetaData().getTables(null, getSchemaPattern(), null, null);
				while (rs.next()) {
					tables.add(rs.getString(3));
				}
			} catch (SQLException e) {
				logger.error(e);
			} finally {
				closeResultSet(rs);
			}
		}
		return tables;
	}

	/**
	 * Gets the columns in a table. The columns are a simple Collection of
	 * DatabaseColumn objects.
	 * 
	 * @param	table	The table to get the columns for.
	 * 
	 * @return A Collection of DatabaseColumns..
	 */
	public Collection<DatabaseColumn> getColumns(String table) {
		Collection<DatabaseColumn> columns = new LinkedList<DatabaseColumn>();
		Connection conn = getConnection();
		if (conn != null) {
			ResultSet rs = null;
			try {
				rs = conn.getMetaData().getColumns(null, null, table, null);
				if (rs == null) return columns;
				while (rs.next()) {
					columns.add(new DatabaseColumn(rs.getString(4), rs.getInt(5)));
				}
			} catch (SQLException e) {
				logger.error(e);
			} finally {
				closeResultSet(rs);
			}
		}
		return columns;
	}

	/**
	 * Closes a result set and handles any exceptions,
	 * 
	 * @param	rs	The result set to close.
	 */
	public void closeResultSet(ResultSet rs) {
		if (rs != null) {
			try {
				rs.close();
			} catch (SQLException e) {
				logger.error(e);
			}
		}
	}
	
	/**
	 * Determines whether the specified column in the specified table is
	 * a primary key.
	 * 
	 * @param	column	The column to check.
	 * @param	table	The table that contains the column.
	 * 
	 * @return	<code>true</code> if the column is a primary key,
	 * 			<code>false</code> otherwise.
	 */
	public boolean isPrimaryKey(String column, String table) {
		boolean retVal = false;
		Connection conn = getConnection();
		if (conn != null && !isOracle()) {
			ResultSet rs = null;
			try {
				DatabaseMetaData dbmd = conn.getMetaData();
				rs = dbmd.getPrimaryKeys(null, null, table);
				while (rs.next()) {
					if (column.equals(rs.getString(4))) {
						retVal = true;
						break;
					}
				}
			} catch (SQLException e) {
				logger.error(e);
			} finally {
				closeResultSet(rs);
			}
		}
		return retVal;
	}

	/**
	 * Determines whether the specified column in the specified table is
	 * a foreign key.
	 * 
	 * @param	column	The column to check.
	 * @param	table	The table that contains the column.
	 * 
	 * @return	<code>true</code> if the column is a foreign key,
	 * 			<code>false</code> otherwise.
	 */
	public boolean isForeignKey(String column, String table) {
		boolean retVal = false;
		Connection conn = getConnection();
		if (conn != null && !isOracle()) {
			ResultSet rs = null;
			try {
				DatabaseMetaData dbmd = conn.getMetaData();
				rs = dbmd.getImportedKeys(null, null, table);
				while (rs.next()) {
					if (column.equals(rs.getString(8))) {
						retVal = true;
						break;
					}
				}
			} catch (SQLException e) {
				logger.error(e);
			} finally {
				closeResultSet(rs);
			}
		}
		return retVal;
	}
	
	/**
	 * Sets the generated keys from the statement that was executed.
	 * 
	 * @param	statement	The statement that was executed.
	 */
	private void setGeneratedKeys(PreparedStatement statement) {
        generatedKeys.clear();
        ResultSet rs = null;
        try {
            rs = statement.getGeneratedKeys();
            if ( rs != null ) {
                while( rs.next() ) {
                    ResultSetMetaData rsm = rs.getMetaData();
                    for( int i = 1; i <= rsm.getColumnCount(); i++ ) {
                        generatedKeys.add(i - 1, new GeneratedKey(rsm.getColumnName(i), rsm.getColumnType(i), rs.getObject(i)));
                    }
                }
            }
            rs.first();
        } catch (SQLException e) {
        	logger.error(e);
        } finally {
        	closeResultSet(rs);
        }
    }
	
	/**
	 * Gets the list of generated keys from the last executed statement.
	 * 
	 * @return	The list of generated keys
	 */
	public List<GeneratedKey> getGeneratedKeys() {
		return Collections.unmodifiableList(generatedKeys);
	}
	
} // QueryAdministrator