/**
 * Copyright (c) 2009, 2011 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.datasource;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Represents data returned from a result set.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class QueryResult {

	private List<DatabaseRow> rows;
	private List<DatabaseColumn> columns;
	
	/**
	 * Creates a new QueryResult based on the result set passed in.
	 * NOTE: The ResultSet should be closed after creating the QueryResult.
	 * This is not handled automatically by the QueryResult.
	 * 
	 * @param	rs	The ResultSet to build the QueryResult from.
	 * 
	 * @throws	SQLException	If an exception occurs while parsing the result set.
	 */
	public QueryResult(ResultSet rs) throws SQLException {
		rows = new ArrayList<DatabaseRow>();
		columns = new ArrayList<DatabaseColumn>();
		setColumns(rs);
		setRows(rs);
	}
	
	/**
     * Sets the rows for this query result.
     *
     * @param   rs  The ResultSet for the query.
     * 
     * @throws	SQLException	If an exception occurs while parsing the result set.
     */
    private void setRows(ResultSet rs) throws SQLException {
        if ( rs != null ) {
            rows.clear();
        	ResultSetMetaData rsmd = rs.getMetaData();
            int colCount = rsmd.getColumnCount();
            while(rs.next()) {
                List<Object> row = new ArrayList<Object>();
                for(int i = 1; i <= colCount; i++) {
                	row.add(i - 1, rs.getObject(i));
                }
                rows.add(new DatabaseRow(rs.getRow(), row));
            }
        }
    }
    
    /**
     * Sets the columns for this query result.
     *
     * @param   rs  The ResultSet for the query.
     * 
     * @throws	SQLException	If an exception occurs while parsing the result set.
     */
    private void setColumns(ResultSet rs) throws SQLException {
    	columns.clear();
        ResultSetMetaData rsm = rs.getMetaData();
        for(int i = 1; i <= rsm.getColumnCount(); i++) {
            columns.add(i - 1, new DatabaseColumn(rsm.getColumnName(i), rsm.getColumnType(i)));
        }
    }
    
    /**
     * Gets the number of columns in this result.
     * 
     * @return	The number of columns in this result.
     */
    public int getColumnCount() {
    	return columns.size();
    }
    
    /**
     * Gets the number of rows in this result.
     * 
     * @return	The number of rows in this result.
     */
    public int getRowCount() {
    	return rows.size();
    }
    
    /**
     * Gets the rows from the result.
     * 
     * @return	The rows in this result.
     */
    public List<DatabaseRow> getRows() {
    	return Collections.unmodifiableList(rows);
    }
    
    /**
     * Gets the columns from the result.
     * 
     * @return	The columns in this result.
     */
    public List<DatabaseColumn> getColumns() {
    	return Collections.unmodifiableList(columns);
    }
	
} // QueryResult