/**
 * Copyright (c) 2009, 2010 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.inflection;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

/**
 * Utility for singularizing and pluralizing words.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class Inflector {
	
	private static Inflector instance;
	
	private List<InflectionRule> plurals;
	private List<InflectionRule> singulars;
	private List<String> uncountables;
	
	/**
	 * Private constructor since this is a singleton. Use the getInstance()
	 * method to get a reference to an Inflector object.
	 */
	private Inflector() {
		plurals = new LinkedList<InflectionRule>();
		singulars = new LinkedList<InflectionRule>();
		uncountables = new LinkedList<String>();
		
		addPlural("$", "s"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("s$", "s"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(ax|test)is$", "$1es"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(octop|vir)us$", "$1i"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(alias|status)$", "$1es"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(bu)s$", "$1ses"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(buffal|tomat)o$", "$1oes"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("([ti])um$", "$1a"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("sis$", "ses"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(?:([^f])fe|([lr])f)$", "$1$2ves"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(hive)$", "$1s"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("([^aeiouy]|qu)y$", "$1ies"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(x|ch|ss|sh)$", "$1es"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(matr|vert|ind)ix|ex$", "$1ices"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("([m|l])ouse$", "$1ice"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("^(ox)$", "$1en"); //$NON-NLS-1$ //$NON-NLS-2$
		addPlural("(quiz)$", "$1zes"); //$NON-NLS-1$ //$NON-NLS-2$

		addSingular("s$", ""); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(n)ews$", "$1ews"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("([ti])a$", "$1um"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("((a)naly|(b)a|(d)iagno|(p)arenthe|(p)rogno|(s)ynop|(t)he)ses$", "$1$2sis"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(^analy)ses$", "$1sis"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("([^f])ves$", "$1fe"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(hive)s$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(tive)s$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("([lr])ves$", "$1f"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("([^aeiouy]|qu)ies$", "$1y"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(s)eries$", "$1eries"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(m)ovies$", "$1ovie"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(x|ch|ss|sh)es$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("([m|l])ice$", "$1ouse"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(bus)es$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(o)es$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(shoe)s$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(cris|ax|test)es$", "$1is"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(octop|vir)i$", "$1us"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(alias|status)es$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("^(ox)en", "$1"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(vert|ind)ices$", "$1ex"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(matr)ices$", "$1ix"); //$NON-NLS-1$ //$NON-NLS-2$
		addSingular("(quiz)zes$", "$1"); //$NON-NLS-1$ //$NON-NLS-2$

		addIrregular("person", "people"); //$NON-NLS-1$ //$NON-NLS-2$
		addIrregular("man", "men"); //$NON-NLS-1$ //$NON-NLS-2$
		addIrregular("child", "children"); //$NON-NLS-1$ //$NON-NLS-2$
		addIrregular("sex", "sexes"); //$NON-NLS-1$ //$NON-NLS-2$
		addIrregular("move", "moves"); //$NON-NLS-1$ //$NON-NLS-2$

		addUncountable("equipment"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("information"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("rice"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("money"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("species"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("series"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("fish"); //$NON-NLS-1$ //$NON-NLS-2$
		addUncountable("sheep"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	/**
	 * Gets a reference to the inflector object.
	 * 
	 * @return	An Inflector object.
	 */
	public static Inflector getInstance() {
		if (instance == null) {
			instance = new Inflector();
		}
		return instance;
	}

	/**
	 * Adds an irregular word to the inflector. Irregular words are
	 * words that have no simple rules to determine the singular or
	 * plural form of them.
	 * 
	 * @param	singular	The singular form of the word.
	 * @param	plural	The plural form of the word.
	 */
	private void addIrregular(String singular, String plural) {
		addPlural("(" + singular.charAt(0) + ")" + singular.substring(1) + "$", "$1" + plural.substring(1)); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		addSingular("(" + plural.charAt(0) + ")" + plural.substring(1) + "$", "$1" + singular.substring(1)); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
	}

	/**
	 * Adds an uncountable word to the inflector. Uncountable words
	 * do not change their form in the singular or plural case.
	 * 
	 * @param	word	The uncountable word to add.
	 */
	private void addUncountable(String word){
		uncountables.add(word.toLowerCase());
	}

	/**
	 * Adds a plural rule to the inflector.
	 * 
	 * @param	rule	A regular expression representing the pattern.
	 * @param	replacement	A regular expression representing the replacement pattern.
	 */
	private void addPlural(String rule, String replacement){
		plurals.add(new InflectionRule(rule, replacement));
	}

	/**
	 * Adds a singular rule to the inflector.
	 * 
	 * @param	rule	A regular expression representing the pattern.
	 * @param	replacement	A regular expression representing the replacement pattern.
	 */
	private void addSingular(String rule, String replacement){
		singulars.add(new InflectionRule(rule, replacement));
	}
	
	/**
	 * Attempts to apply the rules against the specified word.
	 * 
	 * @param	rules	The rules to test against.
	 * @param	word	The word to run the rules against.
	 * 
	 * @return	The word applied by it's first matching rule,
	 * 			or the word itself if the word is uncountable.
	 */
	private String applyRules(List<InflectionRule> rules, String word) {
		String retVal = word;
		if (!uncountables.contains(word.toLowerCase())) {
			ListIterator<InflectionRule> i = rules.listIterator(rules.size());
			while (i.hasPrevious()) {
				InflectionRule rule = i.previous();
				if (rule.isMatch(word)) {
					retVal = rule.apply(word);
					break;
				}
			}
		}
		return retVal;
	}
	
	/**
	 * Pluralizes a word based on the number passed in. If the number
	 * is 1, then the word will be left alone.
	 * 
	 * @param	word	The word to pluralize.
	 * @param	number	The number associated with the number of objects represented by the word.
	 * 
	 * @return	A pluralized version of the word if number is anything but 1, otherwise
	 * 			the word is simply returned.
	 */
	public String pluralize(String word, int number) {
		return number == 1 ? word : pluralize(word);
	}
	
	/**
	 * Converts a word to its plural form.
	 * 
	 * @param	word	The singular form of the word to pluralize.
	 * 
	 * @return	The plural form of the word.
	 */
	public String pluralize(String word) {
		return applyRules(plurals, word);
	}
	
	/**
	 * Converts a word to its singular form.
	 * 
	 * @param	word	The plural form of the word to singularize.
	 * 
	 * @return	The singular form of the word.
	 */
	public String singularize(String word) {
		return applyRules(singulars, word);
	}
	
} // Inflector