/**
 * Copyright (c) 2009, 2010 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.logging;

import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;

/**
 * Provides basic logging functionality.
 *
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class JDORMLogger {

	/**
	 * Provides various levels of logging. DEBUG will show all levels of 
	 * logging, WARN will show only WARN, ERROR, and FATAL, etc.
	 */
	public enum Level {
		ALL,
		DEBUG,
		INFO,
		WARN,
		ERROR,
		FATAL,
		OFF
	};
	
	private static Map<Class<?>, JDORMLogger> loggerMap = new HashMap<Class<?>, JDORMLogger>();
	
	private Class<?> clazz;
	private Level level;
	
	/**
	 * Private constructor since loggers are created using 
	 * the getLogger method.
	 * 
	 * @param	clazz	The class to associate this logger with.
	 */
	private JDORMLogger(Class<?> clazz) {
		this.clazz = clazz;
		this.level = Level.DEBUG;
	}
	
	/**
	 * Gets the JDORMLogger for the specified class. Only one is created
	 * per class.
	 * 
	 * @param	clazz	The class to get the logger for.
	 * 
	 * @return	A JDORMLogger for the specified class.
	 */
	public static JDORMLogger getLogger(Class<?> clazz) {
		JDORMLogger logger = loggerMap.get(clazz);
		if (logger == null) {
			logger = new JDORMLogger(clazz);
			loggerMap.put(clazz, logger);
		}
		return logger;
	}
	
	/**
	 * Sets the current level for the logger.
	 * 
	 * @param	level	The level to log at.
	 */
	public void setLevel(Level level) {
		this.level = level;
	}
	
	/**
	 * Gets the current level for the logger.
	 * 
	 * @return	The current logger lever. (See the Level enum)
	 */
	public Level getLevel() {
		return level;
	}
	
	/**
	 * Logs a fatal error.
	 * 
	 * @param	t	The throwable that caused the error.
	 */
	public void fatal(Throwable t) {
		fatal(null, t);
	}
	
	/**
	 * Logs a fatal error.
	 * 
	 * @param	str	A message to associate with the error.
	 * @param	t	The throwable that caused the error.
	 */
	public void fatal(String str, Throwable t) {
		output(Level.FATAL, str, null);
	}
	
	/**
	 * Logs an error.
	 * 
	 * @param	t	The throwable that caused the error.
	 */
	public void error(Throwable t) {
		error(null, t);
	}
	
	/**
	 * Logs an error.
	 * 
	 * @param	str	A message to associate with the error.
	 * @param	t	The throwable that caused the error.
	 */
	public void error(String str, Throwable t) {
		output(Level.ERROR, str, t);
	}
	
	/**
	 * Logs a warning.
	 * 
	 * @param	t	The throwable that caused the warning.
	 */
	public void warn(Throwable t) {
		warn(null, t);
	}
	
	/**
	 * Logs a warning.
	 * 
	 * @param	str	A message to associate with the warning.
	 * @param	t	The throwable that caused the warning.
	 */
	public void warn(String str, Throwable t) {
		output(Level.WARN, str, t);
	}
	
	/**
	 * Logs an informative message.
	 * 
	 * @param	t	The throwable that caused the informative message.
	 */
	public void info(Throwable t) {
		info(null, t);
	}
	
	/**
	 * Logs an informative message.
	 * 
	 * @param	str	A message to associate with the information.
	 * @param	t	The throwable that caused the informative message.
	 */
	public void info(String str, Throwable t) {
		output(Level.INFO, str, t);
	}
	
	/**
	 * Logs a debug message.
	 * 
	 * @param	t	The throwable that caused the debug message.
	 */
	public void debug(Throwable t) {
		debug(null, t);
	}
	
	/**
	 * Logs a debug message.
	 * 
	 * @param	str	A message to associate with the debug message.
	 * @param	t	The throwable that caused the debug message.
	 */
	public void debug(String str, Throwable t) {
		output(Level.DEBUG, str, t);
	}
	
	/**
	 * Outputs the string and/or throwable based on the current level.
	 * 
	 * @param	level	The level to attempt to log.
	 * @param	str	The string to log (Optional).
	 * @param	t	The throwable to log (Optional).
	 */
	private void output(Level level, String str, Throwable t) {
		if (level.ordinal() >= getLevel().ordinal()) {
			System.err.print(level.toString() + " (" + Calendar.getInstance().getTime().toString() //$NON-NLS-1$
					+ " - " + clazz.getSimpleName() + "): " + str); //$NON-NLS-1$ //$NON-NLS-2$
			if (t != null) {
				t.printStackTrace(System.err);
			}
		}
	}
	
} // JDORMLogger