/**
 * Copyright (c) 2009, 2011 jDORM, http://www.jdorm.com
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package com.jdorm.activerecord;

import static org.junit.Assert.*;

import java.util.Collection;
import java.util.List;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import com.jdorm.datasource.DataSourceType;
import com.jdorm.datasource.DatabaseConnectionManager;
import com.jdorm.datasource.DatabaseDescriptor;
import com.jdorm.datasource.QueryAdministrator;
import com.jdorm.tests.models.Favorite;
import com.jdorm.tests.models.FavoriteUser;
import com.jdorm.tests.models.Post;
import com.jdorm.tests.models.User;

/**
 * Tests the AbstractActiveRecord class.
 * 
 * @author	mbaumbach
 *
 * @version	1.0
 */
public class AbstractActiveRecordTest {

	private static final String DATABASE = "jdorm_test"; //$NON-NLS-1$
	private static final String USER = "root"; //$NON-NLS-1$
	private static final String PASSWORD = ""; //$NON-NLS-1$
	
	private static final String ADMIN_NAME = "Admin"; //$NON-NLS-1$
	private static final String REGULAR_NAME = "User"; //$NON-NLS-1$
	private static final String REGULAR_NAME_CHANGED = "UserChange"; //$NON-NLS-1$
	
	private static final Float ADMIN_SALARY = new Float(70000.25f);
	private static final Float REGULAR_SALARY = new Float(50000.5f);
	private static final Float DEFAULT_FLOAT = new Float(0.0f);
	
	private static QueryAdministrator queryAdministrator;
	
	private User adminUser;
	private User regularUser;
	private User newUser;
	
	private Post adminPost1;
	private Post adminPost2;
	private Post regularPost;
	
	private Favorite adminPost1FavoriteForAdmin;
	private Favorite regularPostFavoriteForAdmin;
	private Favorite regularPostFavoriteForRegular;
	
	/**
	 * @throws java.lang.Exception
	 */
	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		queryAdministrator = new QueryAdministrator(new DatabaseDescriptor(DataSourceType.MySQL, AbstractActiveRecordTest.DATABASE, AbstractActiveRecordTest.USER, AbstractActiveRecordTest.PASSWORD));
		AbstractActiveRecord.establishConnection(queryAdministrator.getDatabaseDescriptor());
	}
	
	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception {
		// Create tables for the various models
		queryAdministrator.execute("CREATE TABLE IF NOT EXISTS users (id INT(11) NOT NULL AUTO_INCREMENT PRIMARY KEY, name VARCHAR(255), email VARCHAR(255), username VARCHAR(255), password VARCHAR(255), admin TINYINT(1), salary FLOAT, created_at DATETIME, updated_at DATETIME)"); //$NON-NLS-1$
		queryAdministrator.execute("CREATE TABLE IF NOT EXISTS posts (id INT(11) NOT NULL AUTO_INCREMENT PRIMARY KEY, title VARCHAR(255), body LONGTEXT, user_id INT(11), created_at DATETIME, updated_at DATETIME)"); //$NON-NLS-1$
		queryAdministrator.execute("CREATE TABLE IF NOT EXISTS favorites (id INT(11) NOT NULL AUTO_INCREMENT PRIMARY KEY, user_id INT(11), post_id INT(11), created_at DATETIME, updated_at DATETIME)"); //$NON-NLS-1$
		
		adminUser = new User();
		adminUser.set("name", ADMIN_NAME); //$NON-NLS-1$
		adminUser.set("email", "admin@jdorm.com"); //$NON-NLS-1$ //$NON-NLS-2$
		adminUser.set("username", "admin"); //$NON-NLS-1$ //$NON-NLS-2$
		adminUser.set("password", "admin_pass"); //$NON-NLS-1$ //$NON-NLS-2$
		adminUser.set("admin", Boolean.TRUE); //$NON-NLS-1$
		adminUser.set("salary", ADMIN_SALARY); //$NON-NLS-1$
		adminUser.save();
		
		regularUser = new User();
		regularUser.set("name", REGULAR_NAME); //$NON-NLS-1$
		regularUser.set("email", "user@jdorm.com"); //$NON-NLS-1$ //$NON-NLS-2$
		regularUser.set("username", "user"); //$NON-NLS-1$ //$NON-NLS-2$
		regularUser.set("password", "user_pass"); //$NON-NLS-1$ //$NON-NLS-2$
		regularUser.set("admin", Boolean.FALSE); //$NON-NLS-1$
		regularUser.set("salary", REGULAR_SALARY); //$NON-NLS-1$
		regularUser.save();
		
		newUser = new User();
		
		adminPost1 = new Post();
		adminPost1.set("title", "Admin Post 1 Title"); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost1.set("body", "Admin Post 1 Body"); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost1.set("user_id", adminUser.get("id"));
		adminPost1.save();
		
		adminPost2 = new Post();
		adminPost2.set("title", "Admin Post 2 Title"); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost2.set("body", "Admin Post 2 Body"); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost2.set("user_id", adminUser.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost2.save();
		
		regularPost = new Post();
		regularPost.set("title", "Regular Post Title"); //$NON-NLS-1$ //$NON-NLS-2$
		regularPost.set("body", "Regular Post Body"); //$NON-NLS-1$ //$NON-NLS-2$
		regularPost.set("user_id", regularUser.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		regularPost.save();
		
		adminPost1FavoriteForAdmin = new Favorite();
		adminPost1FavoriteForAdmin.set("user_id", adminUser.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost1FavoriteForAdmin.set("post_id", adminPost1.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		adminPost1FavoriteForAdmin.save();
		
		regularPostFavoriteForAdmin = new Favorite();
		regularPostFavoriteForAdmin.set("user_id", adminUser.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		regularPostFavoriteForAdmin.set("post_id", regularPost.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		regularPostFavoriteForAdmin.save();

		regularPostFavoriteForRegular = new Favorite();
		regularPostFavoriteForRegular.set("user_id", regularUser.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		regularPostFavoriteForRegular.set("post_id", regularPost.get("id")); //$NON-NLS-1$ //$NON-NLS-2$
		regularPostFavoriteForRegular.save();
	}
	
	/**
	 * @throws java.lang.Exception
	 */
	@After
	public void tearDown() throws Exception {
		// Destroy all of the tables used for testing
		queryAdministrator.execute("DROP TABLE IF EXISTS users"); //$NON-NLS-1$
		queryAdministrator.execute("DROP TABLE IF EXISTS posts"); //$NON-NLS-1$
		queryAdministrator.execute("DROP TABLE IF EXISTS favorites"); //$NON-NLS-1$
	}
	
	/**
	 * @throws java.lang.Exception
	 */
	@AfterClass
	public static void tearDownAfterClass() throws Exception {
		DatabaseConnectionManager.getInstance().closeConnection(queryAdministrator.getDatabaseDescriptor());
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getString(java.lang.String)}.
	 */
	@Test
	public void testGetString() {
		assertEquals(ADMIN_NAME, adminUser.getString("name")); //$NON-NLS-1$
		assertEquals(REGULAR_NAME, regularUser.getString("name")); //$NON-NLS-1$
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getBoolean(java.lang.String)}.
	 */
	@Test
	public void testGetBoolean() {
		assertEquals(Boolean.TRUE, adminUser.getBoolean("admin")); //$NON-NLS-1$
		assertEquals(Boolean.FALSE, regularUser.getBoolean("admin")); //$NON-NLS-1$
		assertEquals(Boolean.FALSE, newUser.getBoolean("admin")); //$NON-NLS-1$
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getFloat(java.lang.String)}.
	 */
	@Test
	public void testGetFloat() {
		assertEquals(ADMIN_SALARY, adminUser.getFloat("salary")); //$NON-NLS-1$
		assertEquals(REGULAR_SALARY, regularUser.getFloat("salary")); //$NON-NLS-1$
		assertEquals(DEFAULT_FLOAT, newUser.getFloat("admin")); //$NON-NLS-1$
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getHasMany(java.lang.Class)}.
	 */
	@Test
	public void testGetHasManyClassOfT() {
		boolean foundAdminPost1 = false;
		boolean foundAdminPost2 = false;
		
		Collection<Post> posts = adminUser.getHasMany(Post.class);
		assertEquals(2, posts.size());
		for (Post post : posts) {
			if (post.equals(adminPost1)) {
				foundAdminPost1 = true;
			} else if (post.equals(adminPost2)) {
				foundAdminPost2 = true;
			}
		}
		
		assertTrue(foundAdminPost1);
		assertTrue(foundAdminPost2);
	}
	
	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getHasManyThrough(java.lang.Class)}.
	 */
	@Test
	public void testGetHasManyThroughClassOfT() {
		boolean foundAdminPost1 = false;
		boolean foundRegularPost = false;
		
		Collection<Post> posts = adminUser.getHasManyThrough(Post.class, Favorite.class);
		assertEquals(2, posts.size());
		for (Post post : posts) {
			if (post.equals(adminPost1)) {
				foundAdminPost1 = true;
			} else if (post.equals(regularPost)) {
				foundRegularPost = true;
			}
		}
		
		assertTrue(foundAdminPost1);
		assertTrue(foundRegularPost);
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getBelongsTo(java.lang.Class)}.
	 */
	@Test
	public void testGetBelongsToClassOfT() {
		assertEquals(regularUser, regularPostFavoriteForRegular.getBelongsTo(User.class));
		assertEquals(regularPost, regularPostFavoriteForRegular.getBelongsTo(Post.class));
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#findAll(java.lang.Class)}.
	 */
	@Test
	public void testFindAllClassOfT() {
		boolean foundAdmin = false;
		boolean foundRegular = false;
		for (User user : AbstractActiveRecord.findAll(User.class)) {
			if (user.equals(adminUser)) {
				foundAdmin = true;
			}
			if (user.equals(regularUser)) {
				foundRegular = true;
			}
		}
		assertTrue(foundAdmin);
		assertTrue(foundRegular);
	}
	
	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#findAll(java.lang.Class, com.jdorm.activerecord.Conditions)}.
	 */
	@Test
	public void testFindAllClassOfTConditions() {
		List<User> users = AbstractActiveRecord.findAll(User.class, Conditions.where("id = ?", regularUser.get("id"))); //$NON-NLS-1$ //$NON-NLS-2$
		assertEquals(1, users.size());
		assertEquals(regularUser, users.get(0));
	}
	
	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#findFirst(java.lang.Class)}.
	 */
	@Test
	public void testFindFirstClassOfT() {
		assertEquals(adminUser, AbstractActiveRecord.findFirst(User.class));
	}
	
	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getFields(java.lang.Class)}.
	 */
	@Test
	public void testGetFieldsClassOfT() {
		List<String> fields = AbstractActiveRecord.getFields(User.class);
		String[] expectedFields = { "id", "name", "email", "username", "password", "admin", "salary", "created_at", "updated_at" };
		assertArrayEquals(expectedFields, fields.toArray(new String[0]));
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#getDatabaseTableName(java.lang.Class)}.
	 */
	@Test
	public void testGetDatabaseTableName() {
		assertEquals("users", AbstractActiveRecord.getDatabaseTableName(User.class)); //$NON-NLS-1$
		assertEquals("posts", AbstractActiveRecord.getDatabaseTableName(Post.class)); //$NON-NLS-1$
		assertEquals("favorites", AbstractActiveRecord.getDatabaseTableName(Favorite.class)); //$NON-NLS-1$
		assertEquals("favorite_users", AbstractActiveRecord.getDatabaseTableName(FavoriteUser.class)); //$NON-NLS-1$
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#isNewRecord()}.
	 */
	@Test
	public void testIsNewRecord() {
		assertTrue(newUser.isNewRecord());
		assertFalse(adminUser.isNewRecord());
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#save()}.
	 */
	@Test
	public void testSave() {
		// Test the update case
		assertEquals(REGULAR_NAME, regularUser.getString("name")); //$NON-NLS-1$
		assertFalse(regularUser.isNewRecord());
		regularUser.set("name", REGULAR_NAME_CHANGED); //$NON-NLS-1$
		assertTrue(regularUser.save());
		assertFalse(regularUser.isNewRecord());
		
		User fetchedUser = AbstractActiveRecord.findFirst(User.class, Conditions.where("id = ?", regularUser.get("id"))); //$NON-NLS-1$ //$NON-NLS-2$
		assertEquals(regularUser, fetchedUser);
		assertEquals(REGULAR_NAME_CHANGED, fetchedUser.getString("name")); //$NON-NLS-1$
		
		// Test the create case
		assertTrue(newUser.isNewRecord());
		newUser.set("name", "Testing"); //$NON-NLS-1$ //$NON-NLS-2$
		assertTrue(newUser.save());
		assertFalse(newUser.isNewRecord());
		Object id = newUser.get("id"); //$NON-NLS-1$
		assertNotSame(0, id);
		
		fetchedUser = AbstractActiveRecord.findFirst(User.class, Conditions.where("id = ?", id)); //$NON-NLS-1$
		assertEquals(newUser, fetchedUser);
		assertEquals("Testing", fetchedUser.getString("name")); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#destroy()}.
	 */
	@Test
	public void testDestroy() {
		assertFalse(newUser.destroy());
		Object id = regularUser.get("id"); //$NON-NLS-1$
		assertTrue(regularUser.destroy());
		assertTrue(regularUser.isNewRecord());
		
		User fetchedUser = AbstractActiveRecord.findFirst(User.class, Conditions.where("id = ?", id)); //$NON-NLS-1$
		assertNull(fetchedUser);
	}

	/**
	 * Test method for {@link com.jdorm.activerecord.AbstractActiveRecord#equals(java.lang.Object)}.
	 */
	@Test
	public void testEqualsObject() {
		assertEquals(newUser, newUser);
		assertNotSame(newUser, adminUser);
		assertEquals(adminUser, adminUser);
		assertNotSame(adminUser, regularUser);
		
		User fetchedUser = AbstractActiveRecord.findFirst(User.class, Conditions.where("id = ?", adminUser.get("id"))); //$NON-NLS-1$ //$NON-NLS-2$
		assertEquals(adminUser, fetchedUser);
	}

} // AbstractActiveRecordTest